from collections import defaultdict
from markupsafe import Markup
from dateutil.relativedelta import relativedelta

from odoo import api, fields, models, Command, _
from odoo.exceptions import AccessDenied, UserError
from odoo.tools import float_compare
from odoo.addons.saas_worker.const import OVH_INVOICING_AK, OVH_INVOICING_AS, OVH_INVOICING_CK
from odoo.addons.saas_worker.util import email_on_exception
from odoo.addons.openerp_enterprise.models.ovh_api import OvhAPI, OvhTooManyLines
from odoo.addons.openerp_enterprise.models.openerp_enterprise import HOSTING_TYPES_SELECTION

import datetime
import logging
import requests

_logger = logging.getLogger(__name__)

ACCOUNT_HOSTING_TYPES_SELECTION = HOSTING_TYPES_SELECTION + [("n/a", "Not Applicable")]


class AccountMove(models.Model):
    _inherit = 'account.move'

    # === Payment widget fields === #
    # can't really put this in standard, as there is no need in a
    # standard case where you only see invoices in active companies.
    invoice_outstanding_credits_debits_widget = fields.Binary(compute_sudo=True)
    invoice_has_outstanding = fields.Boolean(compute_sudo=True)
    invoice_payments_widget = fields.Binary(compute_sudo=True)
    tax_country_id = fields.Many2one(compute_sudo=True)

    hosting = fields.Selection(
        ACCOUNT_HOSTING_TYPES_SELECTION, string="Hosting",
        compute="_compute_hosting", inverse="_inverse_hosting",
    )

    deferred_date_warning = fields.Text(compute='_compute_deferred_date_warning_message')

    @api.depends('invoice_line_ids.hosting')
    def _compute_hosting(self):
        self.hosting = False
        for record in self:
            product_invoice_lines = record.invoice_line_ids.filtered(lambda l: l.display_type == 'product')
            hostings = {hosting for hosting in product_invoice_lines._get_hosting().values()}
            if len(hostings) == 1:
                record.hosting = hostings.pop()

    def _inverse_hosting(self):
        draft_moves = self.filtered(lambda m: m.state == "draft" and m.hosting)
        for move in draft_moves:
            move.invoice_line_ids.filtered(
                lambda l: l.display_type == 'product'
            ).hosting = move.hosting
        # for non-draft move (readonly) force recompute from the move lines
        (self - draft_moves)._compute_hosting()

    @api.depends('invoice_line_ids.product_id', 'invoice_line_ids.deferred_start_date', 'invoice_line_ids.deferred_end_date')
    def _compute_deferred_date_warning_message(self):
        prorated_discount_product_id = self.env.ref("openerp_enterprise.product_prorated_discount").id
        first_year_discount_product_ids = (
            self.env.ref('openerp_enterprise.product_odoo_user_standard_discount').id,
            self.env.ref('openerp_enterprise.product_odoo_user_custom_discount').id,
        )
        self.deferred_date_warning = False
        for move in self:
            warnings = {}
            for line in move.invoice_line_ids.filtered(lambda l: l.product_id):
                if line.product_id.id == prorated_discount_product_id:
                    if not (line.deferred_start_date and line.deferred_end_date):
                        warnings['missing_prorated_deferred'] = \
                            _("You need to fill the deferred start/end dates for line with prorated user discount.")
                elif line.product_id.id in first_year_discount_product_ids:
                    if not (line.deferred_start_date and line.deferred_end_date):
                        warnings['missing_1styeardsc_deferred'] = \
                            _("You need to fill the deferred start/end dates for line with first year discount.")
                    elif line.deferred_end_date > line.deferred_start_date + relativedelta(years=1, days=-1):
                        warnings['invalid_1styeardsc_deferred'] = \
                            _("The deferred start/end dates for line with first year discount must not exceed 1 year.")
                if warnings:
                    self.deferred_date_warning = '\n'.join(warnings.values())

    # =============================================================================================
    # Fix for intercompany invoices.
    # CHE was not able to perform her work with intercompany invoices since their sale orders were
    # not in Odoo S.A. which causes issues with record rules limiting the read&write access on
    # sale orders from other companies.

    def _fix_records_interco_invoices(self):
        """ This method check if any of the record passed in self is an invoice
            for a sale order made in a different company. If so, it returns self with admin rights
            only if the current user has write access on account.move
        """
        for r in self:
            order_ids = r.sudo().mapped('invoice_line_ids.sale_line_ids.order_id')
            if any(order_id.company_id.id != r.company_id.id for order_id in order_ids):
                self.check_access_rule('write')
                self = self.sudo()
        return self

    def button_draft(self):
        return super(AccountMove, self._fix_records_interco_invoices()).button_draft()

    def _invoice_paid_hook(self):
        ret = super(AccountMove, self._fix_records_interco_invoices())._invoice_paid_hook()

        # Copy payment token from payment to subscription when paying on an invoice
        aprs = self.env['account.partial.reconcile'].search([('debit_move_id', 'in', self.line_ids.ids)])
        aprs_by_move_id = defaultdict(aprs.browse)
        for apr in aprs:
            aprs_by_move_id[apr.debit_move_id.move_id.id] += apr
        for move in self:
            # payment_id will be on the reconciled counterpart of our move
            aprs = aprs_by_move_id[move.id]
            payment_token_ids = aprs.credit_move_id.payment_id.payment_token_id
            if payment_token_ids:
                for sub in move.line_ids.sale_line_ids.order_id:
                    if sub.is_subscription and not sub.payment_token_id:
                        sub.sudo().write({'payment_token_id': max(payment_token_ids.ids)})

        return ret


    def button_cancel(self):
        return super(AccountMove, self._fix_records_interco_invoices()).button_cancel()

    def action_post(self):
        if (
            not self.env.context.get('disable_abnormal_invoice_detection', True)  # manual posting by accountant
            and self.sudo().filtered(lambda m: not m.payment_id and m.deferred_date_warning)
        ):
            raise UserError(_("Please fix deferred start/end dates warnings before posting the invoice"))

        res = super(AccountMove, self._fix_records_interco_invoices()).action_post()

        msg = "Pricing Compensation reduced for <a href='#' data-oe-model='sale.order' data-oe-id='%s'>%s</a> (%s)"
        state_field_id = self.env.ref('account.field_account_move__state').id
        for am in self:
            if am.move_type == 'out_refund':  # don't reapply -7pc on Credit Note's confirmation
                continue
            if am.message_ids.sudo().tracking_value_ids.filtered(lambda t: t.field_id.id == state_field_id):
                # in case invoice has been already posted, don't recompute discount
                continue

            # automatically remove 7% of compensation discount line on invoice confirmation
            xtra_disc = self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount')
            lines = am.line_ids.sale_line_ids
            for sol in lines.filtered(lambda l: l.order_id.is_subscription and l.product_id == xtra_disc):
                so = sol.order_id
                pu = sol._reduce_discount_7pc()
                am.message_post(body=Markup(msg % (so.id, so.name, pu)))
        return res

    def write(self, values):
        return super(AccountMove, self._fix_records_interco_invoices()).write(values)
    # ==============================================================================================

    def _message_auto_subscribe_notify(self, partner_ids, *args, **kwargs):
        # Disable auto-sub notifications for invoices.
        # It's not very useful in our case, and will avoid hundreds of spurious
        # notifications for auto-generated ones.
        return

    def _post(self, soft=True):
        # freeze hosting just before posting invoice, so that we can rely on it for Avatax
        self.invoice_line_ids._freeze_hosting()

        posted = super()._post(soft)
        posted.invoice_line_ids.subscription_id.filtered(
            lambda s: s.subscription_state == '4_paused'
        ).resume_subscription()
        return posted


class AccountMoveLine(models.Model):

    _inherit = 'account.move.line'

    # This fails when a record rule forces a join between invoice and invoice line,
    # because the field becomes ambiguous! Need to rewrite read_flat/read_from_database
    # to use generate_order_by() and the query object

    # _order = "period_id DESC,move_id,sequence,id"

    revenue_date_start = fields.Date('Revenue Begin Date', copy=False)
    revenue_date_stop = fields.Date('Revenue End Date', copy=False)
    revenue_monthly = fields.Float("Monthly Revenue", compute='_compute_revenue_monthly', store=True)
    revenue_monthly_sum = fields.Float('Monthly Revenue (EUR)', compute='_compute_report_sum')
    price_subtotal_sum = fields.Float('Amount (EUR)', compute='_compute_report_sum')
    hosting = fields.Selection(ACCOUNT_HOSTING_TYPES_SELECTION, string="Hosting",
                               readonly=True, copy=False, index='btree_not_null')
    subscription_template_id = fields.Many2one(related='subscription_id.sale_order_template_id',
                                               string="Subscription Template", readonly=True)
    product_category_id = fields.Many2one(related='product_id.categ_id', string="Product Category", readonly=True)

    _sql_constraints = [
        ('revenue_dates',
         'check((revenue_date_start is null and revenue_date_stop is null) or '
               '(revenue_date_start is not null and revenue_date_stop is not null '
               ' and revenue_date_start < revenue_date_stop))',
         'Revenue end date must come after revenue begin date'),
    ]

    def _copy_data_extend_business_fields(self, values):
        super()._copy_data_extend_business_fields(values)
        # for move reversals, keep original hosting by default
        values['hosting'] = self.hosting

    def _compute_revenue_monthly(self):
        for invoice_line in self:
            invoice_line.revenue_monthly = 0
            if invoice_line.revenue_date_start:
                revenue = invoice_line.price_unit * invoice_line.quantity * ((100 - invoice_line.discount or 0.0) / 100.0)
                date_start = invoice_line.revenue_date_start
                date_stop = invoice_line.revenue_date_stop

                # NOTE: for boostrapping this in pure SQL, here is a good approximation:
                #   SELECT round((line.revenue_date_stop - line.revenue_date_start) / (365.2425/12))
                # which gives a complete script as:
                #   UPDATE account_invoice_line l SET revenue_monthly =
                #      l.price_subtotal /
                #        round((l.revenue_date_stop - l.revenue_date_start) / (365.2425/12))
                #   WHERE round((l.revenue_date_stop - l.revenue_date_start) / (365.2425/12)) > 0;

                # we use this appromixation instead of exact month differences, as we're rounding up
                # anyways, so that 11.89 months is counted as 12.
                months = round((date_stop - date_start).days / (365.2425 / 12))
                invoice_line.revenue_monthly = revenue / months if months else 0


    def _compute_report_sum(self):
        """ implemented in SQL to exactly match the result of Invoice Analysis """
        self.env.cr.execute("""
           WITH currency_rate AS (
                SELECT
                  r.currency_id,
                  COALESCE(r.company_id, c.id) as company_id,
                  r.rate,
                  r.name AS date_start,
                  (SELECT name FROM res_currency_rate r2
                   WHERE r2.name > r.name AND
                       r2.currency_id = r.currency_id AND
                       (r2.company_id is null or r2.company_id = c.id)
                  ORDER BY r2.name ASC
                  LIMIT 1) AS date_end
                FROM res_currency_rate r
                JOIN res_company c ON (r.company_id is null or r.company_id = c.id)
            )

           SELECT sub.id, sub.price_total / cr.rate AS price_subtotal_sum,
                  sub.revenue_monthly / cr.rate AS revenue_monthly_sum
           FROM (SELECT
                    ail.id,
                    ai.currency_id,
                    ai.invoice_date as date,
                    ai.company_id,
                    (CASE
                      WHEN ai.move_type in ('out_refund','in_invoice')
                      THEN - ail.price_subtotal
                      ELSE ail.price_subtotal
                    END) AS price_total,
                    (CASE
                      WHEN ai.move_type in ('out_refund','in_invoice')
                      THEN - ail.revenue_monthly
                      ELSE ail.revenue_monthly
                    END) AS revenue_monthly
                 FROM account_move_line ail JOIN account_move ai ON ai.id = ail.move_id
                 WHERE ail.id IN %s
                ) sub
           LEFT JOIN currency_rate cr ON
                (cr.currency_id = sub.currency_id AND
                 cr.company_id = sub.company_id AND
                 cr.date_start <= sub.date AND
                 (cr.date_end IS NULL OR cr.date_end > sub.date));

        """, (tuple(self.ids),))
        data = dict((row.pop('id'), row) for row in self.env.cr.dictfetchall())
        for invoice_line in self:
            invoice_line.revenue_monthly_sum = data.get(invoice_line.id)['revenue_monthly_sum']
            invoice_line.price_subtotal_sum = data.get(invoice_line.id)['price_subtotal_sum']

    def _get_hosting(self):
        result = {}
        for move_line in self:
            if move_line.hosting or move_line.parent_state != 'draft':
                result[move_line] = move_line.hosting
            else:
                orders = move_line.subscription_id or move_line.sale_line_ids.order_id
                if orders and not any(o._is_enterprise_plans() for o in orders):
                    # none linked orders are a subscription plan, set the hosting as
                    # "Not Applicable" (ex. partnership)
                    result[move_line] = 'n/a'
                else:
                    result[move_line] = orders[:1].hosting
        return result

    def _freeze_hosting(self):
        draft_unset_move_lines = self.filtered(lambda m: m.parent_state == 'draft' and not m.hosting)
        line_by_hosting = defaultdict(self.env['account.move.line'].browse)
        for move_line, hosting in draft_unset_move_lines._get_hosting().items():
            line_by_hosting[hosting] |= move_line
        for hosting, move_lines in line_by_hosting.items():
            move_lines.hosting = hosting

    @api.model
    def _get_deferred_amounts_by_line_values(self, line):
        """Get deferred amounts for a line
        :param line: dict (account report) or account.move.line
        """
        values = super()._get_deferred_amounts_by_line_values(line)
        values['hosting'] = line['hosting'] or False
        if 'commission_plan_id' in line:  # 'commission_plan_id' field does not exists on move line
            values['commission_plan_id'] = line['commission_plan_id'] or False
        return values


class AccountTax(models.Model):

    _inherit = 'account.tax'

    # Protect "python" taxes from non-admin modifications. It's not clear we need them at all, btw.
    def _create(self, vals_list):
        if any(vals.get('python_compute') or vals.get('python_applicable') for vals in vals_list) and not self.env.is_system():
            raise AccessDenied("Python taxes are managed by administrators")
        return super()._create(vals_list)

    def _write(self, vals):
        if (vals.get('python_compute') or vals.get('python_applicable')) and not self.env.is_system():
            raise AccessDenied("Python taxes are managed by administrators")
        return super()._write(vals)


class AccountInvoiceReport(models.Model):

    _inherit = 'account.invoice.report'

    revenue_monthly = fields.Float('Monthly Revenue', readonly=True)
    revenue_date_start = fields.Date('Revenue Begin Date', readonly=True)
    revenue_date_stop = fields.Date('Revenue End Date', readonly=True)
    hosting = fields.Selection(
        string="Hosting", readonly=True,
        selection=ACCOUNT_HOSTING_TYPES_SELECTION, selection_add=[('n/a', 'Not Applicable')],
    )
    has_referrer = fields.Boolean("Referrer set/not set", readonly=True)
    subscription_template_id = fields.Many2one('sale.order.template', string="Subscription Template", readonly=True)

    def _from(self):
        return super()._from() + """
            LEFT JOIN sale_order AS sub ON (line.subscription_id = sub.id)
        """

    def _select(self):
        return super(AccountInvoiceReport, self)._select() + \
            """, CASE
                     WHEN move.move_type in ('out_refund', 'in_invoice')
                        THEN -1 * line.revenue_monthly
                        ELSE line.revenue_monthly
                    END AS revenue_monthly,
               line.revenue_date_start,
               line.revenue_date_stop,
               line.hosting,
               bool(move.referrer_id) AS has_referrer,
               sub.sale_order_template_id AS subscription_template_id
            """


class account_payment(models.Model):
    _inherit = "account.payment"

    def _compute_destination_account_id(self):
        handled_ids = []
        for rec in self:
            if rec.reconciled_invoice_ids:
                if rec.company_id != rec.reconciled_invoice_ids[0].company_id:
                    if rec.partner_type == 'customer':
                        default_account = rec.env['ir.property'].with_company(rec.company_id).get('property_account_receivable_id', 'res.partner')
                        rec.destination_account_id = default_account.id
                        handled_ids.append(rec.id)
                        return
        return super(account_payment, self.filtered(lambda x: x.id not in handled_ids))._compute_destination_account_id()


class AccountMoveOvh(models.Model):
    _inherit = 'account.move'

    def _ovh_guess_account(self, line):
        ACCOUNTS = {
            'server': self.env.ref('openerp_enterprise.account_account_server_date').id,  # 7770 - 612141 Serveur Data
            'voip': self.env.ref('openerp_enterprise.account_account_phone').id,  # 650 - 612100 Téléphone
            'website': self.env.ref('openerp_enterprise.account_account_website').id,  # 932 - 612160 Site internet
        }

        description = line.get('description', '') + ' ' + line.get('domain', '')

        if '003281139' in description:
            return ACCOUNTS['voip']

        WEBSITE_DESC = [
            "demande de renouvellement",
            "demande d'enregistrement",
            "domain name",
            "redir-",
        ]
        if any(desc in description for desc in WEBSITE_DESC):
            return ACCOUNTS['website']

        SERVER_DESC = [
            "ip-",
            "Snapshots Public Cloud",
            "region gra1",
            "Infrastructure",
            "Advance",
        ]
        if any(desc in description for desc in SERVER_DESC):
            return ACCOUNTS['server']

        fallback = ACCOUNTS['server']  # fallback on Serveur Data
        _logger.info('Cannot match account for line %s, fallback on %s', line, fallback)
        return fallback

    @api.model
    @email_on_exception()
    def _cron_sync_ovh_bills_and_refunds(self, days_in_past=30):
        tax_21_service_intra = self.env.ref('openerp_enterprise.ovh_tax_intra_s_21').ids  # [46, ]

        ovh = OvhAPI('https://eu.api.ovh.com', OVH_INVOICING_AK, OVH_INVOICING_AS, OVH_INVOICING_CK)

        from_date = str((datetime.date.today() - datetime.timedelta(days=days_in_past)))
        today = str(datetime.date.today())

        bills = ovh.getBills(from_date, today)
        refunds = ovh.getRefunds(from_date, today)
        existing = self.search([('ref', 'in', bills + refunds)]).mapped('ref')

        fn = {
            'bill': ovh.getBill,
            'billDetails': ovh.getBillDetails,
            'refund': ovh.getRefund,
            'refundDetails': ovh.getRefundDetails,
        }

        for doc in (set(bills + refunds) - set(existing)):
            _type = doc in bills and 'bill' or 'refund'
            _logger.info('importing %s %s', _type, doc)

            bill = fn[_type](doc)

            am_values = {
                'move_type': _type == 'bill' and 'in_invoice' or 'in_refund',
                'ref': doc,
                'partner_id': self.env.ref('openerp_enterprise.ovh_bill_partner_id').id,
                'invoice_date': bill['date'].split('T')[0],
                'invoice_line_ids': [],
            }

            try:
                tooManyLines = False
                lines = list(fn[_type + 'Details'](doc))
            except OvhTooManyLines:
                tooManyLines = True
                # shortcut to avoid to make 200+ request
                lines = [{
                    'quantity': 1,
                    'unitPrice': bill['priceWithoutTax'],
                    'description': 'To many lines ... ip-',
                }]

            if _type == 'bill' and len(lines) <= 40 and not tooManyLines:
                for line in lines:
                    line_desc = f"[{line['domain']}] {line['description']}"
                    if line['periodStart']:
                        line_desc += f" ({line['periodStart']} - {line['periodEnd']}"

                    if line['unitPrice']['value'] < 0.01:
                        line_desc += f" [{line['quantity']} x {line['unitPrice']['value']}]"
                        line['unitPrice']['value'] = line['totalPrice']['value']
                        line['quantity'] = 1

                    am_values['invoice_line_ids'].append(Command.create({
                        'name': line_desc,
                        'quantity': line['quantity'],
                        'price_unit': line['unitPrice']['value'],
                        'tax_ids': [(6, 0, tax_21_service_intra)],
                        'ref': line[_type + 'DetailId'],
                        'account_id': self._ovh_guess_account(line),
                    }))
            else:
                total_by_account = defaultdict(float)
                for line in lines:
                    total_by_account[self._ovh_guess_account(line)] += float(line['quantity']) * line['unitPrice']['value']
                for account, total in total_by_account.items():
                    name = "Renouvellement {date_split[1]}/{date_split[0]}".format(date_split=am_values['invoice_date'].split('-'))
                    am_values['invoice_line_ids'].append(Command.create({
                        'name': _type == 'bill' and name or "Avoir OVH",
                        'quantity': 1,
                        'price_unit': total * (_type == 'bill' and 1 or -1),
                        'tax_ids': [(6, 0, tax_21_service_intra)],
                        'account_id': account,
                    }))

            new_bill = self.create(am_values)
            assert float_compare(new_bill.amount_total, abs(bill['priceWithTax']['value']), 2) == 0, \
                f"{bill['priceWithTax']['value']} != {new_bill.amount_total}"
            self.env.cr.commit()

            try:
                pdf = requests.get(bill['pdfUrl'], timeout=60)
                self.env['ir.attachment'].create({
                    'res_id': new_bill.id,
                    'res_model': self._name,
                    'raw': pdf.content,
                    'name': doc + '.pdf'
                })
            except requests.exceptions.Timeout:
                new_bill.message_post(
                    body="Invoice cannot be downloaded from: %s" % bill['pdfUrl'],
                    message_type='comment',
                    subtype_xmlid='mail.mt_note',
                )
            self.env.cr.commit()

        return
